#!/usr/bin/env python

"""
Handlers for a person for whom scheduling is performed.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.data import get_uri
from imiptools.dates import format_datetime
from imiptools.handlers import Handler
from imiptools.handlers.common import CommonFreebusy
from imiptools.period import Period, replace_overlapping
from imiptools.profile import Preferences

class PersonHandler(Handler):

    "Handling mechanisms specific to people."

    def _record(self, from_organiser=True, queue=False, cancel=False):

        """
        Record details from the current object given a message originating
        from an organiser if 'from_organiser' is set to a true value, queuing a
        request if 'queue' is set to a true value, or cancelling an event if
        'cancel' is set to a true value.
        """

        # Obtain valid organiser and attendee details.

        oa = self.require_organiser_and_attendees(from_organiser)
        if not oa:
            return False

        (organiser, organiser_attr), attendees = organiser_item, attendees = oa

        # Handle notifications and invitations.

        if from_organiser:

            # Process each attendee separately.

            for attendee, attendee_attr in attendees.items():

                if not self.have_new_object(attendee):
                    continue

                # Set the complete event or an additional occurrence.

                self.store.set_event(attendee, self.uid, self.recurrenceid, self.obj.to_node())

                # Remove additional recurrences if handling a complete event.

                if not self.recurrenceid:
                    self.store.remove_recurrences(attendee, self.uid)

                # Queue any request.

                if queue:
                    self.store.queue_request(attendee, self.uid, self.recurrenceid)
                elif cancel:
                    self.store.cancel_event(attendee, self.uid, self.recurrenceid)

                    # No return message will occur to update the free/busy
                    # information, so this is done here.

                    freebusy = self.store.get_freebusy(attendee)
                    self.remove_from_freebusy(freebusy)

                    self.store.set_freebusy(attendee, freebusy)

                    if self.publisher:
                        self.publisher.set_freebusy(attendee, freebusy)

                self.update_freebusy_from_organiser(attendee, organiser_item)

        # As organiser, update attendance from valid attendees.

        else:
            if self.merge_attendance(attendees, organiser):
                self.update_freebusy_from_attendees(organiser, attendees)

        return True

    def _record_freebusy(self, from_organiser=True):

        """
        Record free/busy information for a message originating from an organiser
        if 'from_organiser' is set to a true value.
        """

        if from_organiser:
            organiser_item = self.require_organiser(from_organiser)
            if not organiser_item:
                return

            senders = [organiser_item]
        else:
            oa = self.require_organiser_and_attendees(from_organiser)
            if not oa:
                return

            organiser_item, attendees = oa
            senders = attendees.items()

            if not senders:
                return

        freebusy = []

        for value in self.obj.get_values("FREEBUSY") or []:
            if not isinstance(value, list):
                value = [value]
            for v in value:
                try:
                    start, end = v.split("/", 1)
                    freebusy.append(Period(start, end))
                except ValueError:
                    pass

        dtstart = format_datetime(self.obj.get_utc_datetime("DTSTART"))
        dtend = format_datetime(self.obj.get_utc_datetime("DTEND"))
        user = get_uri(self.recipient)

        for sender, sender_attr in senders:
            stored_freebusy = self.store.get_freebusy_for_other(user, sender)
            replace_overlapping(stored_freebusy, Period(dtstart, dtend), freebusy)
            self.store.set_freebusy_for_other(user, stored_freebusy, sender)

class Event(PersonHandler):

    "An event handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("An addition to an event has been received.", link=False)

    def cancel(self):

        "Queue a cancellation of any active event."

        self._record(from_organiser=True, queue=False, cancel=True)
        return self.wrap("A cancellation has been received.", link=False)

    def counter(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("A counter proposal has been received.", link=False)

    def declinecounter(self):

        # NOTE: Queue a suggested modification to any active event.

        return self.wrap("A declining counter proposal has been received.", link=False)

    def publish(self):

        "Register details of any relevant event."

        self._record(from_organiser=True, queue=False)
        return self.wrap("Details of an event have been received.")

    def refresh(self):

        "Generate details of any active event."

        # NOTE: Return event details if configured to do so.

        return self.wrap("A request for updated event details has been received.")

    def reply(self):

        "Record replies and notify the recipient."

        self._record(from_organiser=False, queue=False)
        return self.wrap("A reply has been received.")

    def request(self):

        "Hold requests and notify the recipient."

        self._record(from_organiser=True, queue=True)
        return self.wrap("A request has been received.")

class Freebusy(PersonHandler, CommonFreebusy):

    "A free/busy handler."

    def publish(self):

        "Register free/busy information."

        self._record_freebusy(from_organiser=True)

        # Produce a message if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_messages") == "notify":
            return self.wrap("A free/busy update has been received.", link=False)

    def reply(self):

        "Record replies and notify the recipient."

        self._record_freebusy(from_organiser=False)

        # Produce a message if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_messages") == "notify":
            return self.wrap("A reply to a free/busy request has been received.", link=False)

    def request(self):

        """
        Respond to a request by preparing a reply containing free/busy
        information for each indicated attendee.
        """

        # Produce a reply if configured to do so.

        preferences = Preferences(get_uri(self.recipient))
        if preferences.get("freebusy_sharing") == "share":
            return CommonFreebusy.request(self)

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ]

# vim: tabstop=4 expandtab shiftwidth=4
